{
  "type": "record",
  "name": "EditableMLFeatureTableProperties",
  "namespace": "com.linkedin.pegasus2avro.ml.metadata",
  "doc": "Properties associated with a MLFeatureTable editable from the ui",
  "fields": [
    {
      "name": "description",
      "type": [
        "null",
        "string"
      ],
      "doc": "Documentation of the MLFeatureTable",
      "default": null,
      "Searchable": {
        "fieldName": "editedDescription",
        "fieldType": "TEXT"
      }
    }
  ],
  "Aspect": {
    "name": "editableMlFeatureTableProperties"
  }
}